#include "pcrio.h"

#ifdef DEBUG_VER_
#include <stdio.h>
#endif // DEBUG_VER_

extern int errno;

////////////////////////////////////////////////
////  Set variable to default values   /////////
////////////////////////////////////////////////
PcrIO::PcrIO(QObject *parent, const char *name)
  : QObject(parent, name)
{
  fd = -1;
  bytesToWrite = 0;
  bufferReadLen = 0;
  timerId = 0;
  bzero(bufferRead, PCRIO_BUFFERSIZE);
  bzero(bufferWrite, PCRIO_BUFFERSIZE);
}


PcrIO::~PcrIO()
{
}


/////////////////////////////////////////////////
///////////  Set highPriority ///////////////////
/////////////////////////////////////////////////
bool PcrIO::highPriority () const{ return true; }




/////////////////////////////////////////////////////////
// Open serial port to PCR1000 and initialize baudrate //
// If sucessful, it sends connected signals            //
// Then, start timer to check every interval if there  //
// is a data coming in from PCR1000                    //
/////////////////////////////////////////////////////////
bool PcrIO::Open(const char *device, tcflag_t baudrate)
{
  // does not check if device is already opened
  // potential bug here ...

  // open serial device
  fd = open(device, O_RDWR | O_NONBLOCK );
  if(fd < 0){
    perror(device);
    return false;
  }

  // set termio attributes
  tcgetattr(fd, &tio);
  bzero( &tio, sizeof(struct termios));
  cfsetispeed( &tio, baudrate);
  cfsetospeed( &tio, baudrate);
  
  tio.c_cflag |= (CLOCAL | CREAD);
  tio.c_cflag &= ~HUPCL ;
  tio.c_cflag &= ~PARENB ;
  tio.c_cflag &= ~CSTOPB ;
  tio.c_cflag &= ~CSIZE;
  tio.c_cflag |=  CS8;
  
  tio.c_lflag &= ~(ECHO | ECHOCTL);
  
  tio.c_oflag &= ~OPOST;
  tio.c_oflag &= ~(VTDLY|FFDLY|BSDLY|TABDLY|CRDLY|NLDLY);
  
  tcsetattr(fd, TCSANOW, &tio);

  emit connected();


#ifdef DEBUG_VER_
  fprintf(stderr, "PcrIO::Open - %s successful\n", device);
#endif // DEBUG_VER_
  
  
  timerId = startTimer(10);

  return true;
}



//////////////////////////////////////////
// Close serial port and kill timer     //
// if sucessful, it emits closed signal //
//////////////////////////////////////////
bool PcrIO::Close()
{
  close(fd);
  killTimer( timerId );
  fd = -1;

  emit closed();

  return true;
}



/////////////////////////////////////////////////////////
// sendMessageSlot waits for data from external source //
// and forward the message to PCR1000                  //
/////////////////////////////////////////////////////////
void PcrIO::sendMessageSlot(const char *data, int len)
{
  // check file descriptor
  if(fd < 0) return;

  memcpy(bufferWrite+bytesToWrite, data, len);
  bytesToWrite += len;

#ifdef DEBUG_VER_
  fprintf(stderr, "PcrIO::sendMessageSlot - bytesToWrite = %d\n%s\n",
				bytesToWrite, data);
#endif // DEBUG_VER_

  ReadWrite();
}



////////////////////////////////////////////////////////////////////////
// IsTokenComplete checks if data received from PCR1000 has completed //
// at least one token                                                 //
// first --> points to first character of the token                   //
// tokenlen --> token len                                             //
////////////////////////////////////////////////////////////////////////
bool PcrIO::IsTokenComplete(const char *data, int len, int *first, int *tokenlen)
{
  int i;

  // reply from PCR1000 always begins with 'I' for normal inquery
  // and 'N' for bandscope responses
  for(i = 0; i < len; i++)
	if(data[i] == 'I' && (len-i)>=4){
	  *first = i;
	  *tokenlen = 4;
	  return true;

	}else if(data[i] == 'N' && (len-i)>=37){
	  
	  *first = i;
	  *tokenlen = 37;
	  return true;
	}

  return false;
}




////////////////////////////////////////////////////////////////////////
// timerEvent checks to see if there is data waiting to be read/write //
////////////////////////////////////////////////////////////////////////
void PcrIO::timerEvent( QTimerEvent *e){

  ReadWrite();
}



///////////////////////////////////
// ReadWrite does IO operations  //
///////////////////////////////////
bool PcrIO::ReadWrite()
{
  int n;
  int pos, len;
  struct timeval tv;

  tv.tv_usec = 0;
  tv.tv_sec  = 0;

  FD_ZERO(&rset);
  FD_ZERO(&wset);
  
  FD_SET(fd, &rset);
  
  if(bytesToWrite > 0)
	FD_SET(fd, &wset);
  
  // check if there is data waiting
  select(fd+1, &rset, &wset, NULL, &tv);
  
  if(FD_ISSET(fd, &rset)){

	if((n = read(fd, bufferRead+bufferReadLen, PCRIO_BUFFERSIZE-bufferReadLen))<0){
	  if(errno != EAGAIN)
		perror("read error on socket");
	}else if(n == 0){
	  
	  // get disconnected somehow
	  // killTimer( timerId );
	  emit disconnected();
#ifdef DEBUG_VER_
        fprintf(stderr, "PcrIO::ReadWrite - emit disconnected()\n");
#endif // DEBUG_VER_


	}else{
	 
	  // read successful
	  bufferReadLen += n;
	  if(IsTokenComplete(bufferRead, bufferReadLen, &pos, &len)){
		
		// send message
		bzero(mesg, PCRIO_BUFFERSIZE);
		memcpy(mesg, bufferRead+pos, len);
		emit radioMessage(mesg, len);
#ifdef DEBUG_VER_
		fprintf(stderr, "PcrIO::ReadWrite - emit radioMessage(%s, %d)\n", mesg,len);
#endif // DEBUG_VER_
		
		bufferReadLen -= (pos+len);
		memmove(bufferRead, bufferRead+pos+len, bufferReadLen);
	  }
	  
	}
  }

    
  if(FD_ISSET(fd, &wset)){
	if((n = write(fd, bufferWrite, bytesToWrite)) < 0){
	  if(errno != EAGAIN)
		perror("write error to socket");
	}else{
	  bytesToWrite -= n;
	  if(n < bytesToWrite){
		memmove( bufferWrite, bufferWrite + n, bytesToWrite);
	  }else{
		emit messageSent();
	  }
	}
  }
  
  return true;
}
